 /*
     TopSort
        SJSU Students
      
     Program will create a tree and a memory manager, and from the args will create a BST.
     The args will also contain variouse comands for the tree.

     @param Args :  l1 r1 l2 r2 l3 r3 .... ln rn
     
   */

public class TopSort 
{
   private static int C = 100;
   public static int time = 0;
   public static Node[] finishedStack = null;
   
    /**
      Main
      
    graph driver
    
    Takes in a input from the command line arguments then coverts it into a proper 
    graph. and then performs a topological sort on the graph. then the largest node 
    is checked if it has a cycle and it is the output.

    @param Args :  l1 r1 l2 r2 l3 r3 .... ln rn
     
   */
   public static void main(String[] args) 
   {
      int largest = FindLargest(args);
      int[] preNodes = ConvertArray(args);
      
      int[] count  = countingSort(preNodes);
      int vertNum = numVert(count);
      int paths = preNodes.length/2;
      
      Graph g =new Graph(vertNum);
     
      for(int i = 0; i < count.length; i++)
      {
         if(count[i] > 0)
         {
            Node node = new Node(i);
            g.addVertex(node);
         }
      }
      
      Node nodeA = null;
      
      for(int i = 0; i < preNodes.length-1; i+=2)
      {
         nodeA = g.returnNode(preNodes[i]);
         nodeA.addAdjNode(g.returnNode(preNodes[i+1]));
      }

      TopologicalSort(g);
      
      cycleCheck();
      
   }
   
    /**
     TopologicalSort(Graph graph)
     
     Performs a topological sort on the graph and creates a list of the nodes sorted.
      
     @param graph : graph of the input data.
     
    */ 
   
   public static void TopologicalSort(Graph graph)
   {
      DFS(graph);
   }
   
     /**
     DFS(Graph graph)
     
    Performs a depth first search on the graph based on the input data.
      
     @param graph : graph of the input data.
     
    */ 
   
   public static void DFS(Graph graph)
   {
      for(int i = 0; i < graph.nodes.length; i++)
      {
         graph.nodes[i].color = Node.Color.White;
      }
      time = 0;
      for(int i =0; i < graph.nodes.length;i++)
      {
         if(graph.nodes[i].color == Node.Color.White)
         {
            DFSVisit(graph, graph.nodes[i]);
         }
      }
   }
   
     /*
     DFSvisit(Graph graph)
     
     Method to properly sort the graph in the depth first search 
      
     @param graph : graph of the input data.
     @param node : current node being visited
     
    */ 
   
   public static void DFSVisit(Graph graph, Node node)
   {
      time++;
      node.time = time;
      node.color = Node.Color.Gray;
      if(node.adjNodes != null)
      {
         for(int i = 0; i < node.adjNodes.length; i++)
         {
            if(node.adjNodes[i].color == Node.Color.White)
            {
               node.adjNodes[i].parent = node;
               DFSVisit(graph, node.adjNodes[i]);
            }
         }
      }
      node.color = Node.Color.Black;
      time++;
      node.finished = time;
      addToStack(node);
   }
   
     /*
     AddToStack(Node node)
     
     Adds a node to the stack to be output and cycle checked
      
     @param node : node to be added to the stack
     
    */ 
   
   public static void addToStack(Node node)
   {
      if(finishedStack == null)
      {
         finishedStack = new Node[1];
      }
      else
      {
         Node[] tmp = finishedStack;
         finishedStack = new Node[tmp.length + 1];
         copy(tmp);
      }
      finishedStack[finishedStack.length-1] = node;
   }
   
     /*
     outputStack()
     
     Output the stack that was created after the topological sort
     on the graph
     
    */ 
   
   public static void outputStack()
   {
     for(int i = 0; i < finishedStack.length; i++)
     {
        System.out.print(finishedStack[i].position + " ");
     }
     System.out.println();
   }
   
     /**
     CycleCheck()
     
     Check the stack created from the topological sort to
     se if it has a cycle or not, and then output the resutls.
     
    */
   
   public static void cycleCheck()
   {
      Node[] cycle = null;
      int pos = 0;
      int largest = 0;
      int indexL = 0;
      
      for(int i = 0; i < finishedStack.length;i++)
      {
         if(finishedStack[i].position > largest)
         {
            indexL = i;
            largest = finishedStack[i].position;
         }
      }
      
      int index2 = indexL;
      
      cycle = add(cycle, finishedStack[index2]);
      index2++;
      
      while(index2< finishedStack.length 
                  && finishedStack[index2].edge != true 
                      && finishedStack[index2].position != largest)
      {
         cycle = add(cycle, finishedStack[index2]);
         index2++;
      }
     
      for(int i = cycle.length-1; i >= 0; i--)
      {
         System.out.print(cycle[i].position + " ");
      }
      System.out.println();
      if(cycle[cycle.length-1].parent != null 
              && cycle[0].position == cycle[cycle.length-1].parent.position)
      
      {
         System.out.println("Has Cycle");
      }
      else
      {
         System.out.println("No Cycle");
      }
   }
   
     /**
     add(node[] array, Node node)
     
     Add node to array helper method
      
     @param array : array of nodes
     @param node :  node to add to array
     
    */ 
   
   static Node[] add(Node[] array, Node node)
   {
      if(array == null)
      {
         array = new Node[1];
         array[0] = node;
      }
      else
      {
        Node[] tmp = array;
        array = new Node[tmp.length + 1];
      
      
         for(int i = 0; i < tmp.length; i++)
         {
            array[i] = tmp[i];
         }
         
          array[tmp.length] = node;
      }
      
      return array;
   }
   
     /**
     copy(Node[] aray)
     
     Add array to stack helper method
      
     @param array : array of nodes to be added to he stack
     
    */ 
   
   static void copy(Node[] array)
   {
      for(int i =0; i < array.length;i++)
      {
         finishedStack[i] = array[i];
      }
   }
   
   static int[] countingSort(int[] array)
   {
      int[] count = new int[C];
      
      for(int i = 0; i < C;i++)
      {
         for(int j = 0; j < array.length;j++)
         {
            if(i == array[j])
            {
               count[i]++; 
            }
         } 
     }
      return count;
   }
   
     /**
     numVert(int[] array)
     
     Determines the number of vertexes in the graph
      
     @param array : array of the input data.
     
    */ 
   static int numVert(int[] array)
   {
      int vNum = 0;
       for(int i = 0; i < array.length;i++)
      {
         if(array[i] > 0)
         {
            vNum++;
         }
      }
       return vNum;
   }

     /**
     CovertArray(String[] array)
     
     Coverts input array from string to int
      
     @param array : string array of the input data.
     @return array : int array of the input data
     
    */ 
   static int[] ConvertArray(String[] array)
   {
      int[] nodes = new int[array.length];
      
      for(int i = 0; i < array.length; i++)
      {
         nodes[i] = Integer.parseInt(array[i]);
      }
      
      return nodes;
   }
   
     /**
     FindLargest(String[] array)
     
     Finds the largest node in the graph
      
     @return largest : index of largest
     
    */ 
  static int FindLargest(String[] array)
   {
      int largest = 0;
      
      for(int i = 0; i < array.length;i++)
      {
         if(largest < Integer.parseInt(array[i]))
         {
            largest = Integer.parseInt(array[i]);
         }
      }
      
      return largest;
   }
   
    /**
     outputGraph(Graph graph)
     
     Outputs the graph to user, proof of grapg creation
      
     @param graph : graph of the input data.
     
    */ 
  
   public static void outputGraph(Graph g)
   {
        for(int i = 0; i < g.nodes.length; i++)
      {
         System.out.print(g.nodes[i].position+ " ");
         
         if(g.nodes[i].edge == true)
         {
            System.out.print("E");
         }
         
         if( g.nodes[i].adjNodes != null && g.nodes[i].adjNodes.length > 0)
         {
            for(int j = 0 ; j < g.nodes[i].adjNodes.length; j++)
            {
               System.out.print(g.nodes[i].adjNodes[j] + "");
            }
         }
         System.out.println();
      }
   }
}
